#include <assert.h>
#include <stdio.h>
#include "backend.h"
#include "exceptions.h"
#include "native-utils.h"
#include "subtype.h"

/*
 * Note: if titanium ever grows threads, each
 * thread must have its own copy of this global.
 */

struct ExceptionState *exception_state[MAX_BOX_PROCS];

int find_handler( const struct ExceptionState * const state,
		  const struct class_header * const catchers[] )
{
    if (!state->handling) {
	int candidate;
	for (candidate = 0; catchers[ candidate ]; ++candidate)
	    if (instanceOfClass( state->value, catchers[ candidate ] ))
		return candidate;
    }
    
    return NO_HANDLER;
}


void toss( const GP_instance exception )
{
    CHECK_NULL_GLOBAL( exception );
    
    if (!FILE_STATIC_STATIC_REF(exception_state)) {
	fputs( "fatal error: uncaught exception\n", stderr );
	abort();
    }
    
#ifdef linux
#ifdef __GLIBC__
    assert( FILE_STATIC_STATIC_REF(exception_state)->handler[0].__jmpbuf[ JB_BP ] );
    assert( FILE_STATIC_STATIC_REF(exception_state)->handler[0].__jmpbuf[ JB_SP ] );
    assert( FILE_STATIC_STATIC_REF(exception_state)->handler[0].__jmpbuf[ JB_PC ] );
#else  /* Linux, old libc */
    assert( FILE_STATIC_STATIC_REF(exception_state)->handler[0].__bp );
    assert( FILE_STATIC_STATIC_REF(exception_state)->handler[0].__sp );
    assert( FILE_STATIC_STATIC_REF(exception_state)->handler[0].__pc );
#endif /* Linux, old libc */
#else  /* not Linux */
    /* add other system-specific tests for reasonable jmp_buf */
#endif /* not Linux */
    
    FILE_STATIC_STATIC_REF(exception_state)->value = exception;
    LONGJMP( FILE_STATIC_STATIC_REF(exception_state)->handler, 1 );
}
