tlib_dir    = $(top_srcdir)/tlib
tlib_java   = $(tlib_dir)/tlib.java
tlib_base   = libtlib-$(backend)
tlib_g	    = $(tlib_base)-g.a
tlib_O      = $(tlib_base)-O.a
tlib_incdir = $(datadir)/tlib-include
tlib_kind   = $(patsubst $(tlib_base)%.a,%,$@)

tlib_both = $(tlib_g) $(tlib_O)
 tlibs = $(tlib_both)
#tlibs = $(tlib_O)

tc            = $(top_builddir)/tc
tcbuild_exec  = tcbuild
tcbuild	      = $(top_builddir)/tcbuild/$(tcbuild_exec)
tc_cache      = tc-cache$(tlib_kind)
tc_gen        = tc-gen$(tlib_kind)
tc_gen_seq    = ../sequential/tc-gen$(tlib_kind)
tc_obj        = tc-obj$(tlib_kind)

tcbuild_flags_common = --backend $(backend) --library --outfile $@ --rebuild-tlib $(tlib_kind) --fail-keep
# tlib-slim: minimizes file space usage (best default for non-compiler-developers)
tcbuild_flags_slim = $(tcbuild_flags_common) --skip-tc --cache-dir "" --generate-dir $(tc_gen_seq) --object-dir $(tc_obj) --keep-generate-dir --nokeep-object-dir
# tlib-fast: uses a per-backend compilation cache to reduce C compilation time on recompiles of tlib
tcbuild_flags_fast = $(tcbuild_flags_common) --skip-tc --cache-dir $(tc_cache) --generate-dir $(tc_gen_seq) --object-dir $(tc_obj) --keep-generate-dir --nokeep-object-dir
# tlib-full: naive legacy behavior, every backend gets a separate tc compile and keeps cache & gen dirs
tcbuild_flags_full = $(tcbuild_flags_common) --cache-dir $(tc_cache) --generate-dir $(tc_gen) --keep

ifeq "$(wildcard tc-cache-*)" ""
# default is slim
tlib_strategy = slim
else
# unless we see a cache, then use fast
tlib_strategy = fast
endif

tcbuild_flags = $(tcbuild_flags_$(tlib_strategy))

tlib-slim tlib-slim-O tlib-slim-g: force
	@$(MAKE) tlib_strategy=slim $(subst -slim,,$@) || exit 1

tlib-fast tlib-fast-O tlib-fast-g: force
	@$(MAKE) tlib_strategy=fast $(subst -fast,,$@) || exit 1

tlib-full tlib-full-O tlib-full-g: force
	@$(MAKE) tlib_strategy=full $(subst -full,,$@) || exit 1

force:

.PHONY: force tlib tlib-O tlib-g tlib_O tlib_g

#$(tc): force
#	$(MAKE) -C $(@D) $(@F) all-am

#$(tcbuild): force
#	$(MAKE) -C $(@D)

tlib-g tlib_g: $(tlib_g)
tlib-O tlib_O: $(tlib_O)
 tlib: tlib_g tlib_O
#tlib: tlib_O

$(tlib_both): $(tlib_java)
	@if test ! -x "$(tc)" -o ! -x "$(tcbuild)"; then \
	 	echo "ERROR: tried to build a tlib without tc/tcbuild translator" ; \
                echo "       you must complete a top-level 'make all' before building tlibs" ; \
		exit 1 ; \
	fi
	$(tcbuild) $(tcbuild_flags) $<
	@if test "$@" = "libtlib-sequential-O.a" ; then \
		$(MAKE) -C $(top_builddir)/tcbuild tcdemangle ; \
	fi

install-tlib: $(tlibs) force
	$(mkinstalldirs) $(DESTDIR)$(libdir)
	$(INSTALL_DATA) $(tlib_O) $(DESTDIR)$(libdir)/$(tlib_O)
	$(RANLIB) $(DESTDIR)$(libdir)/$(tlib_O)
	$(INSTALL_DATA) $(tlib_g) $(DESTDIR)$(libdir)/$(tlib_g)
	$(RANLIB) $(DESTDIR)$(libdir)/$(tlib_g)

uninstall-tlib: force
	rm -f $(DESTDIR)$(libdir)/$(tlib_g) $(DESTDIR)$(libdir)/$(tlib_O)

install-tlib-headers: $(tlibs) force
	$(mkinstalldirs) $(DESTDIR)$(tlib_incdir)
	for header in tc-gen-O/*.h; do \
	$(INSTALL_DATA) "$$header" "$(DESTDIR)$(tlib_incdir)/"`echo "$$header" | awk -F/ '{print $$2;}'` || exit 1 ; \
	done

uninstall-tlib-headers: force
	for header in $(DESTDIR)$(tlib_incdir)/*.h; do \
	  rm -f $$header ; \
        done

CLEANFILES = $(tlib_both)

clean-tc-dirs: force
	rm -f $(CLEANFILES)
	rm -rf tc-cache-O tc-cache-g tc-gen-O tc-gen-g tc-obj-O tc-obj-g

$(top_builddir)/tlib/tlib.mk: $(top_srcdir)/tlib/tlib.mk.in
	gmake -C $(top_builddir)/tlib tlib.mk
